#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "graphics.h"
#include "keyboard.h"
#include "objects.h"
#include "tile.h"

object objlist[4096];
int nobj=0;

int blocks(int tile)
{ return (tile && (tile<9 || tile>12) && tile!=26 && !(tile>=31 && tile<=33));
}

int spritecollide(int x, int y, int x2, int y2)
{ if(x>x2+14*256 || x+14*256<x2 || y>y2+15*256 || y+15*256<y2) return 0;
  if(y<y2) return -1;
  return 1;
}

int blocked(int x, int y)
{ int n;
  int i;

  if(n=blocks(backmap[(x/256+1)/16][(y/256+1)/16])) return n;
  if(n=blocks(backmap[(x/256+14)/16][(y/256+1)/16])) return n;
  if(n=blocks(backmap[(x/256+1)/16][(y/256+14)/16])) return n;
  if(n=blocks(backmap[(x/256+14)/16][(y/256+14)/16])) return n;
  for(i=0; i<nobj; i++)
  { if(!objlist[i].blocking) continue;
    if(spritecollide(objlist[i].data[0], objlist[i].data[1], x, y) &&
       objlist[i].blocking(objlist+i, x, y)) 
      return 1;
  }
  return 0;
}

void blitplayer(object *self, int x, int y)
{ char s[256];

  sprintf(s, "score: %-6d   keys: %c%c%c", self->data[12],
                                           self->data[10]?'r':' ',
                                           self->data[9]?'g':' ',
                                           self->data[11]?'b':' ');
  displaytext(vscreen, 0, 0, s);

  if(self->data[8])         /* dead */
    blittranstile(vscreen, obj, 9+PLAYER, self->data[0]/256-x,
                                           self->data[1]/256-y);
  else if(DOWN)             /* ducking */
    blittranstile(vscreen, obj, 7+PLAYER, self->data[0]/256-x,
                                          self->data[1]/256-y);
  else if(self->data[3]>0)  /* falling */
    blittranstile(vscreen, obj, 8+PLAYER, self->data[0]/256-x,
                                          self->data[1]/256-y);
  else if(self->data[3]<0)  /* jumping */
    blittranstile(vscreen, obj, 6+PLAYER, self->data[0]/256-x,
                                          self->data[1]/256-y);
  else if(self->data[5]<3)  /* walking */
    blittranstile(vscreen, obj, self->data[6]+self->data[5]*2+PLAYER,
                                self->data[0]/256-x,
                                self->data[1]/256-y);
  else if(self->data[5]==3)
    blittranstile(vscreen, obj, self->data[6]+PLAYER,
                                self->data[0]/256-x,
                                self->data[1]/256-y);
}

void blitblob(object *self, int x, int y)
{ if(self->data[7]<0)       /* squished */
    blittranstile(vscreen, obj, BLOB+7, self->data[0]/256-x,
                                        self->data[1]/256-y);
  else if(self->data[7])    /* being squished */
    blittranstile(vscreen, obj, BLOB+6, self->data[0]/256-x,
                                        self->data[1]/256-y);
  else if(self->data[2]<5)  /* blobbing */
    blittranstile(vscreen, obj, BLOB+self->data[2], self->data[0]/256-x,
                                                    self->data[1]/256-y);
  else
    blittranstile(vscreen, obj, BLOB+8-self->data[2], self->data[0]/256-x,
                                                      self->data[1]/256-y);
}

void blitcoin(object *self, int x, int y)
{ blittranstile(vscreen, obj, COIN+self->data[2], self->data[0]/256-x,
                                                  self->data[1]/256-y);
}

void blitspike(object *self, int x, int y)
{ if(self->data[2]<4)
    blittranstile(vscreen, obj, SPIKE+self->data[2], self->data[0]/256-x,
                                                     self->data[1]/256-y);
  else
    blittranstile(vscreen, obj, SPIKE+8-self->data[2], self->data[0]/256-x,
                                                       self->data[1]/256-y);
}

void blitkey(object *self, int x, int y)
{ blittranstile(vscreen, obj, GREENKEY+self->data[2], self->data[0]/256-x,
                                                      self->data[1]/256-y);
}

void blitlock(object *self, int x, int y)
{ blittranstile(vscreen, obj, GREENLOCK+self->data[2], self->data[0]/256-x,
                                                       self->data[1]/256-y);
}

void blitbreaker(object *self, int x, int y)
{ blittranstile(vscreen, obj, BREAKER+self->data[2]/16, self->data[0]/256-x,
                                                        self->data[1]/256-y);
}

void blitspringboard(object *self, int x, int y)
{ blittranstile(vscreen, obj, SPRINGBOARD+self->data[2], self->data[0]/256-x,
                                                         self->data[1]/256-y);
}

void blitmovingplat(object *self, int x, int y)
{ blittranstile(vscreen, obj, MOVINGPLAT, self->data[0]/256-x,
                                          self->data[1]/256-y);
}

void blitportcullis(object *self, int x, int y)
{ blittranstile(vscreen, obj, PORTCULLIS+self->data[2], self->data[0]/256-x,
                                                        self->data[1]/256-y);
}

void advframeplayer(object *self)
{ int waterlogged;

  waterlogged=                                   /* in the water? */
    (backmap[(self->data[0]/256+8)/16][(self->data[1]/256+8)/16]>=WATER &&
     backmap[(self->data[0]/256+8)/16][(self->data[1]/256+8)/16]<=WATER+3);

  if(blocked(self->data[0]+self->data[2], self->data[1]+256) || waterlogged)
  { if(self->data[2]>0) self->data[2]-=4;        /* dampen x velocity */
    if(self->data[2]<0) self->data[2]+=4;
    if(self->data[2]>-4 && self->data[2]<4) self->data[2]=0;
    if(!DOWN && !self->data[8]) self->data[2]+=(RIGHT-LEFT)*5; /* move x */
    if(!UP || self->data[8]) self->data[7]=0;    /* not jumping */
  }
  if(!DOWN && !self->data[8]) self->data[2]+=(RIGHT-LEFT)*3; /* move x more */
  if(RIGHT && !self->data[8]) self->data[6]=0;   /* select frame */
  if(LEFT && !self->data[8]) self->data[6]=1;
  else if(self->data[6]>1) self->data[6]-=2;
  if(self->data[2]>256) self->data[2]=256;       /* limit x velocity */
  if(self->data[2]<-256) self->data[2]=-256;

  if(waterlogged) self->data[2]/=4;              /* slower in water */
  self->data[0]+=self->data[2];                  /* move x */
  if(blocked(self->data[0], self->data[1]))      /* if now in wall.. */
    self->data[0]-=self->data[2], self->data[2]=0; /* move back and stop */
  if(waterlogged) self->data[2]*=4;              /* end of slower in water */

  if(!blocked(self->data[0], self->data[1]+256)) self->data[3]+=4; /* fall */
  if(UP && !self->data[8] && !self->data[7] &&   /* jump attempt */
     blocked(self->data[0], self->data[1]+256))
  { self->data[3]=-256-abs(self->data[2])/4;     /* uppity! */
    self->data[7]=1;
  }
  else if(UP && !self->data[8] && waterlogged)   /* swim attempt */
  { self->data[3]=-128;                          /* uppity! */
    self->data[7]=1;
  }

  if(self->data[3]>512) self->data[3]=512;       /* limit y velocity */
  if(self->data[3]<-512) self->data[3]=-512;

  if(waterlogged) self->data[3]/=4;              /* slower in water */
  self->data[1]+=self->data[3];                  /* move z */
  if(blocked(self->data[0], self->data[1]))      /* if now in wall */
    self->data[1]-=self->data[3], self->data[3]=0; /* move back and stop */
  if(waterlogged) self->data[3]*=4;              /* end of slower in water */

  if(self->data[2] && ++self->data[4]>512/abs(self->data[2]))
  { if(++self->data[5]>3) self->data[5]=0;       /* if moving, move feet */
    self->data[4]=0;
  }
}

void advframecoin(object *self)
{ if(spritecollide(self->data[0], self->data[1], objlist[0].data[0],
                                                 objlist[0].data[1]))
  { self->remove=1;                              /* player collected me! */
    objlist[0].data[12]+=4;
  }
  if(++self->data[3]>7)                          /* next frame */
  { self->data[2]++;
    if(self->data[2]>7) self->data[2]=0;
    self->data[3]=0;
  }
}

void advframeblob(object *self)
{
  if(self->data[7]<0)                            /* death timer .. */
  { if(--self->data[7]<-20)
    { self->remove=1;                            /* run out? kill me! */
      objlist[0].data[12]+=2;                    /* and reward the player! */
    }
    return;
  }
   
  if(spritecollide(self->data[0], self->data[1], objlist[0].data[0],
                                                 objlist[0].data[1]))
  { if((!self->data[7] && self->data[1]-objlist[0].data[1]<12*256) ||
       objlist[0].data[3]<=0)      /* if player collides from below */
      objlist[0].data[8]=1;        /* kill him! */
    else
    { self->data[7]=self->data[1]-objlist[0].data[1]; /* being squished */
      if(self->data[7]<2*256)                      /* completely squished? */
      { self->data[7]=-1;                          /* begin death timer */
        objlist[0].data[3]=-objlist[0].data[3];    /* bounce player back */
        return;
      }
   }
  }
  else self->data[7]=0;   /* no longer being squished */
  if(++self->data[3]>10)  /* next frame */
  { self->data[2]++;
    if(self->data[2]>5) self->data[2]=0;
    self->data[3]=0;
  }
  if(!blocked(self->data[0], self->data[1]+256)) self->data[5]+=4; /* fall */
  if(self->data[5]>512) self->data[5]=512;      /* limit y velocity */
  self->data[1]+=self->data[5];                 /* move y */
  if(blocked(self->data[0], self->data[1]))     /* hit ground? */
    self->data[1]-=self->data[5], self->data[5]=0; /* stop! */
  if(!self->data[4] && blocked(self->data[0], self->data[1]+256))
    self->data[4]=-64;         /* if on ground and not moving x, begin */
  self->data[0]+=self->data[4];   /* move x */
  if(blocked(self->data[0], self->data[1])) /* hit wall? */
  { self->data[0]-=self->data[4];           /* backstep */ 
    if(blocked(self->data[0], self->data[1]+256)) /* still hit wall? */
      self->data[4]=-self->data[4];   /* no, reverse x motion */
    else
      self->data[4]=0;                /* yes, stop x motion */
  }
}

void advframespike(object *self)
{ if(++self->data[3]>8)               /* next frame */
  { if(++self->data[2]>7) self->data[2]=0;
    self->data[3]=0;
  }
  if(self->data[1]-objlist[0].data[1]>12*256 && /* player touching tip? */
     spritecollide(self->data[0], self->data[1], objlist[0].data[0],
                                                 objlist[0].data[1]))
    objlist[0].data[8]=1;             /* kill him! */
}

void advframekey(object *self)
{ if(spritecollide(self->data[0], self->data[1], objlist[0].data[0],
                                                 objlist[0].data[1]))
  { objlist[0].data[9+self->data[2]]=1; /* if touched by player, add to */
    self->remove=1;                 /* player inventory and remove self */
  }
}

void advframemovingplat(object *self)
{ self->data[0]+=self->data[2];    /* move */
  self->data[1]+=self->data[3];
  if(blocked(self->data[0], self->data[1])) /* bounce off wall */
  { self->data[0]-=self->data[2];
    self->data[1]-=self->data[3];
    self->data[2]=-self->data[2];
    self->data[3]=-self->data[3];
  }
  if(spritecollide(self->data[0], self->data[1], objlist[0].data[0],
                                                 objlist[0].data[1]) &&
     objlist[0].data[1]<self->data[1]+3*256) /* player on or being pushed? */
  { objlist[0].data[0]+=self->data[2];       /* make him move */
    objlist[0].data[1]+=self->data[3];
    if(blocked(objlist[0].data[0], objlist[0].data[1])) /* player stuck? */
    { objlist[0].data[0]-=self->data[2];    /* put him back */
      objlist[0].data[1]-=self->data[3];
      if(blocked(objlist[0].data[0], objlist[0].data[1])) /* still stuck? */
      { self->data[2]=-self->data[2];       /* reverse platform movement */
        self->data[3]=-self->data[3];
      }
    }
  }
}

void advframespringboard(object *self)
{ if(self->data[3])                    /* spring release timer active? */
  { self->data[3]-=16;                 /* decrease timer */
    self->data[2]=self->data[3]/64;    /* pick a frame */
    if(blocked(objlist[0].data[0], objlist[0].data[1])) /* player on top? */
    { objlist[0].data[1]=self->data[1]+self->data[2]*4*256-14*256; /* move */
      if(!self->data[2])               /* in final position? */
        objlist[0].data[3]=-768;       /* send player flying! */
    }
  }
  else if(!spritecollide(self->data[0], self->data[1], objlist[0].data[0],
                                                       objlist[0].data[1]))
    self->data[3]=self->data[2]*64; /* if player not on, spring up */
  else if(spritecollide(self->data[0], self->data[1], objlist[0].data[0],
                                                      objlist[0].data[1]) &&
          objlist[0].data[1]>self->data[1]+self->data[2]*4*256-15*256)
    if(++self->data[2]>3)           /* if player on, push spring down */
      self->data[3]=256;        /* and if fully down, start release timer */
}

void advframeportcullis(object *self)
{ if(spritecollide(self->data[0], self->data[1], objlist[0].data[0],
                                                 objlist[0].data[1]))
    self->data[2]=1;  /* player underneath?  start lowering.. */
  else if(self->data[2] && ++self->data[3]>4)
  { self->data[3]=0;  /* player no longer underneath?  finish lowering */
    if(++self->data[2]>7) self->data[2]=7;
  }
}

int blockinglock(object *self, int x, int y)
{ if(y<self->data[1]-14*256) return 0;   /* can go */
  if(objlist[0].data[9+self->data[2]] && /* if touched by player with key */
     spritecollide(self->data[0], self->data[1], objlist[0].data[0],
                                                 objlist[0].data[1]))
    self->remove=1;                      /* remove self */
  return !self->remove;
}

int blockingbreaker(object *self, int x, int y)
{ if(y>self->data[1]+3*256-self->data[2]*256/32) return 0; /* can go */
  if(y>self->data[1]-14*256)           /* someone's walking on me?! */
  { self->data[2]++;                   /* weaken! */
    if(self->data[2]>128) self->remove=1; /* if strength is gone, remove! */
    return 1;
  }
  return 0;
}

int blockingmovingplat(object *self, int x, int y)
{ if((x==self->data[0] && y==self->data[1]) ||
     y>self->data[1]+3*256 || y<self->data[1]-14*256)
    return 0;            /* no, you're not walking on me */
 return 1;              /* shut up!  yes you are! */
}

int blockingspringboard(object *self, int x, int y)
{ if(y>self->data[1]-14*256+self->data[2]*4*256) return 1; /* blocked! */
  return 0;                                             /* not! */
}

int blockingportcullis(object *self, int x, int y)
{ if(self->data[2]<2) return 0; /* low enough for ya? */
  return 1;
}

void emptyfunction(object *self)
{}

void *zcalloc(int a, int b)
{ void *c;

  c=calloc(a, b);
  memset(c, 0, a*b);
  return c;
}

void createobject(int type, int x, int y)
{ objlist[nobj].type=type;
  objlist[nobj].remove=0;
  objlist[nobj].blocking=0;
  switch(type)
  { case PLAYER:
      objlist[nobj].data=zcalloc(4, 13);
      objlist[nobj].blit=blitplayer;
      objlist[nobj].advframe=advframeplayer;
      break;
    case BLOB:
      objlist[nobj].data=zcalloc(4, 8);
      objlist[nobj].blit=blitblob;
      objlist[nobj].advframe=advframeblob;
      break;
    case COIN:
      objlist[nobj].data=zcalloc(4, 4);
      objlist[nobj].blit=blitcoin;
      objlist[nobj].advframe=advframecoin;
      break;
    case SPIKE:
      objlist[nobj].data=zcalloc(4, 4);
      objlist[nobj].blit=blitspike;
      objlist[nobj].advframe=advframespike;
      break;
    case GREENKEY:
      objlist[nobj].data=zcalloc(4, 3);
      objlist[nobj].blit=blitkey;
      objlist[nobj].advframe=advframekey;
      break;
    case REDKEY:
      objlist[nobj].data=zcalloc(4, 3);
      objlist[nobj].blit=blitkey;
      objlist[nobj].advframe=advframekey;
      objlist[nobj].data[2]=1;
      break;
    case BLUEKEY:
      objlist[nobj].data=zcalloc(4, 3);
      objlist[nobj].blit=blitkey;
      objlist[nobj].advframe=advframekey;
      objlist[nobj].data[2]=2;
      break;
    case GREENLOCK:
      objlist[nobj].data=zcalloc(4, 3);
      objlist[nobj].blit=blitlock;
      objlist[nobj].advframe=emptyfunction;
      objlist[nobj].blocking=blockinglock;
      objlist[nobj].data[2]=0;
      break;
    case REDLOCK:
      objlist[nobj].data=zcalloc(4, 3);
      objlist[nobj].blit=blitlock;
      objlist[nobj].advframe=emptyfunction;
      objlist[nobj].blocking=blockinglock;
      objlist[nobj].data[2]=1;
      break;
    case BLUELOCK:
      objlist[nobj].data=zcalloc(4, 3);
      objlist[nobj].blit=blitlock;
      objlist[nobj].advframe=emptyfunction;
      objlist[nobj].blocking=blockinglock;
      objlist[nobj].data[2]=2;
      break;
    case BREAKER:
      objlist[nobj].data=zcalloc(4, 3);
      objlist[nobj].blit=blitbreaker;
      objlist[nobj].advframe=emptyfunction;
      objlist[nobj].blocking=blockingbreaker;
      break;
    case SPRINGBOARD:
      objlist[nobj].data=zcalloc(4, 4);
      objlist[nobj].blit=blitspringboard;
      objlist[nobj].advframe=advframespringboard;
      objlist[nobj].blocking=blockingspringboard;
      break;
    case MOVINGPLAT:
      objlist[nobj].data=zcalloc(4, 4);
      objlist[nobj].blit=blitmovingplat;
      objlist[nobj].advframe=advframemovingplat;
      objlist[nobj].blocking=blockingmovingplat;
      if(rand()%2)
        objlist[nobj].data[2]=64;
      else
        objlist[nobj].data[3]=64;
      break;
    case PORTCULLIS:
      objlist[nobj].data=zcalloc(4, 114);
      objlist[nobj].blit=blitportcullis;
      objlist[nobj].advframe=advframeportcullis;
      objlist[nobj].blocking=blockingportcullis;
      objlist[nobj].data[2]=0;
      break;
  }
  objlist[nobj].data[0]=x*256;
  objlist[nobj].data[1]=y*256;
  nobj++;
}

void removeobjects()
{ int i;

  for(i=0; i<nobj; i++)
    if(objlist[i].remove)
      objlist[i]=objlist[--nobj];
}
